import java.util.Vector;
import java.awt.*;

abstract class genericSplineCurve {
		
	protected java.util.Vector controlPoints;
	protected java.util.Vector intermediatePoints;
	protected int nbControlPoints;
	protected int nbPoints;
	protected int selectedControlPoint;
	protected int nbPointsPerSegment;	
	protected boolean reCompute;	
	protected boolean closed;
  protected int nbPointsToCopyToCloseCurve;

	
	public genericSplineCurve() {
		
		super();
		nbControlPoints = 0;
		nbPoints = 0;
		nbPointsPerSegment = 10;
		selectedControlPoint = -1;
		reCompute = true;
		closed = false;
		
		controlPoints = new Vector();
		intermediatePoints = new Vector();
		nbPointsToCopyToCloseCurve = 1;
		
	}

	
/*********************************/
/* Gestion des points de control */
/*********************************/

	public void reset() {
	
		nbControlPoints = 0;
		nbPoints = 0;
		nbPointsPerSegment = 10;
		selectedControlPoint = -1;
		reCompute = true;
		closed = false;
		
		controlPoints = new Vector();
		intermediatePoints = new Vector();
		
	}	
	
		
	public boolean addControlPoint(double x, double y) {
	
		boolean reClose;
		
		if (selectedControlPoint==-1) {
		
			reClose = closed;
			
			if (closed)
					closed(false);
		
	    controlPoints.addElement(new coord2D(x,y));
	    
	    if (reClose)
	    	closed(true);
	    
	    nbControlPoints++;
	    reCompute = true;	
		}
	
		return (selectedControlPoint==-1);
	
	}
	

	public void setControlPoints(Vector v) {
	
		nbControlPoints = v.size();	    
		controlPoints = (Vector) v.clone();
		if (closed) {
			closed = false;
			closed(true);
		}
			
		reCompute = true;
	
	}

	
	public int getNbControlPoints() {
		
			return nbControlPoints;
			
	}

	
	public boolean selectCloserControlPoint(double x, double y, double threshold) {

		double d;
		double px,py;

		selectedControlPoint = -1;
		for (int i=0;i<nbControlPoints && selectedControlPoint==-1;i++) {
	  	  px = ((coord2D)(controlPoints.elementAt(i))).x;
	    	py = ((coord2D)(controlPoints.elementAt(i))).y;
	    	d = Math.sqrt((x-px)*(x-px)+(y-py)*(y-py));
	    	if (d<threshold)
					selectedControlPoint = i;
		}
  
		return (selectedControlPoint!=-1);
		
	}
	

	public int getNumSelectedControlPoint() {
		
		return selectedControlPoint;
		
	}


	public boolean selectControlPoint(int n) {
	
		if (nbControlPoints>=n && n>=0)
	  	  selectedControlPoint = n;
  
		return  (nbControlPoints>=n && n>=0);
	
	}
	
	
	public void deselectControlPoint() {
	
		selectedControlPoint = -1;
	
	}	
	
	
	public boolean moveSelectedControlPoint(double x, double y) {
			
		boolean reClose;
		
		if (selectedControlPoint!=-1) {
			reClose = closed;			
			if (closed)
				closed(false);
	    controlPoints.setElementAt(new coord2D(x,y),selectedControlPoint);
	    reCompute = true;
	    if (reClose)
	    	closed(true);
		}
	
		return (selectedControlPoint!=-1);
	
	}


	public boolean removeSelectedControlPoint() {
	
		boolean reClose;
		
		if (selectedControlPoint!=-1) {
			reClose = closed;			
			if (closed)
				closed(false);
	  	for(int i=selectedControlPoint;i<nbControlPoints-1;i++)
				controlPoints.setElementAt((coord2D)(controlPoints.elementAt(i+1)),i);
	    deselectControlPoint();
	    nbControlPoints--;
	    controlPoints.setSize(nbControlPoints);
	    reCompute = true;
	    if (reClose)
	    	closed(true);
		}
	
		return (selectedControlPoint!=-1);
	
	}	
	

	public Vector getControlPoints() {

		Vector v;
		int x,y;

		v = new Vector();

		for(int i=0;i<nbControlPoints;i++) 	    
	  	  v.addElement(controlPoints.elementAt(i));

		return v;

	}
	
	
	public int[] getControlLineX() {

		int theX[];
		
		theX = new int[nbControlPoints];
		
		for(int i=0;i<nbControlPoints;i++)
	    	theX[i] = (int)(((coord2D)(controlPoints.elementAt(i))).x);
	    	
		return theX;
	
	}
	
	
	public int[] getControlLineY() {

		int theY[];
		
		theY = new int[nbControlPoints];
		
		for(int i=0;i<nbControlPoints;i++) 
	    	theY[i] = (int)(((coord2D)(controlPoints.elementAt(i))).y);
	
		return theY;
	
	}
	
	
/**********************************/
/* Gestion de la courbe de spline */
/**********************************/
	
	public void smoothForBezierCurve() {
	
		boolean reClose;
		double x,y;
		
		for(int i=3;i<nbControlPoints;i+=3) {

			if (i!=0 && i!=nbControlPoints-1) {
				
				reClose = closed;			
				if (closed)
					closed(false);
				
	  		x = 0.5 * ( ((coord2D)(controlPoints.elementAt(i-1))).x + ((coord2D)(controlPoints.elementAt(i+1))).x );
				y = 0.5 * ( ((coord2D)(controlPoints.elementAt(i-1))).y + ((coord2D)(controlPoints.elementAt(i+1))).y );

				controlPoints.setElementAt(new coord2D(x,y),i);
	    
	    	if (reClose)
	    		closed(true);
			}
	    	
		}
		
		reCompute = true;

	}
	
	
	public void closed(boolean closed) {
		
		if (closed && !this.closed) {
			nbControlPoints+=nbPointsToCopyToCloseCurve;
			for(int i=0;i<nbPointsToCopyToCloseCurve;i++)
				controlPoints.addElement(controlPoints.elementAt(i));
		}
		else if (!closed && this.closed) {
			nbControlPoints-=nbPointsToCopyToCloseCurve;
			controlPoints.setSize(nbControlPoints);						
		}
			
		this.closed = closed;
		reCompute = true;
		
	}
	
	
	public void increasePrecision() {
	
		nbPointsPerSegment++;
		reCompute = true;
	}

	
	public void decreasePrecision() {
	
		if (nbPointsPerSegment>1) {
			nbPointsPerSegment--;
			reCompute = true;
		}
		
	}	
	
	public int getNbCurvePoints() {
		
		return nbPoints;
		
	}
	
	public int[] getSplineCurveX() {

		int theX[];

		if (reCompute) {
	  	  computeSplineCurve();
		    reCompute = false;	    
		}         

		theX = new int[nbPoints];

		for(int i=0;i<nbPoints;i++) 
	    	theX[i] = (int)(((coord2D)(intermediatePoints.elementAt(i))).x);

		return theX;	
	}

	
	public int[] getSplineCurveY() {

		int theY[];

		if (reCompute) {
	  	  computeSplineCurve();
		    reCompute = false;	    
		}         

		theY = new int[nbPoints];
		
		for(int i=0;i<nbPoints;i++)
	    	theY[i] = (int)(((coord2D)(intermediatePoints.elementAt(i))).y);
	    	
		return theY;	
	}
	
	
	abstract public void computeSplineCurve();
	
	
}